// 🐦 Flutter imports:
import 'package:flutter/material.dart';

// 📦 Package imports:
import 'package:flutter_localizations/flutter_localizations.dart';
import 'package:go_router/go_router.dart';
import 'package:provider/provider.dart';
import 'package:provider/single_child_widget.dart';
// 🌎 Project imports:
import 'package:verttice_ui/verttice_ui.dart';

typedef ThemedRoute = GoRoute;
typedef ThemedRouteBase = RouteBase;
typedef ThemedRouteState = GoRouterState;

typedef ThemedRouter = GoRouter;

class ThemedApp extends StatelessWidget {
  final String title;
  final bool debugShowCheckedModeBanner;
  final bool showPerformanceOverlay;
  final Widget errorPageWidget;
  final List<ThemedRouteBase> routes;
  final List<SingleChildWidget>? providers;
  final List<NavigatorObserver>? observers;
  final Widget Function(
    BuildContext context,
    ThemedRouteState state,
    Widget child,
  )? pageBuilder;
  final GlobalKey<NavigatorState>? rootNavigatorKey;
  final GlobalKey<NavigatorState>? nestedNavigatorKey;
  final String initialLocation;
  const ThemedApp({
    Key? key,
    required this.title,
    required this.routes,
    this.pageBuilder,
    this.providers,
    this.initialLocation = '/',
    this.rootNavigatorKey,
    this.nestedNavigatorKey,
    this.errorPageWidget = const ThemedMessage(
      message: "404 - Unknown Page",
    ),
    this.observers,
    this.debugShowCheckedModeBanner = true,
    this.showPerformanceOverlay = false,
  }) : super(key: key);

  GoRouter get _router => GoRouter(
        navigatorKey: rootNavigatorKey,
        initialLocation: initialLocation,
        routes: routes,
        observers: observers,
        errorPageBuilder: (context, state) => MaterialPage<void>(
          key: state.pageKey,
          child: errorPageWidget,
        ),
      );

  @override
  Widget build(BuildContext context) {
    return MultiProvider(
      providers: [
        ChangeNotifierProvider(create: (_) => ThemedLocaleState()),
        ChangeNotifierProvider(create: (_) => VertticeUIState()),
        ...?providers,
      ],
      child: Consumer<ThemedLocaleState>(builder: (context, themedLocale, _) {
        return Consumer<VertticeUIState>(
          builder: (context, value, _) {
            return MaterialApp.router(
              highContrastTheme: ThemeData(),
              locale: themedLocalToLocale[themedLocale.locale],
              localizationsDelegates: const [
                GlobalMaterialLocalizations.delegate,
                GlobalWidgetsLocalizations.delegate,
                GlobalCupertinoLocalizations.delegate,
                GlobalWidgetsLocalizations.delegate,
              ],
              supportedLocales: VertticeUI
                  .instance.localeManager.appSupportLocale
                  .map((e) => themedLocalToLocale[e]!)
                  .toList(),
              routeInformationProvider: _router.routeInformationProvider,
              routeInformationParser: _router.routeInformationParser,
              routerDelegate: _router.routerDelegate,
              color: value.themedColors.accentColor,
              theme: ThemeData(
                useMaterial3: true,
                primaryColor: value.themedColors.accentColor,
                colorScheme: ColorScheme.fromSeed(
                  seedColor: value.themedColors.accentColor,
                  brightness:
                      value.isDarkTheme ? Brightness.dark : Brightness.light,
                ),
              ),
              darkTheme: ThemeData(
                useMaterial3: true,
                primaryColor: value.themedColors.accentColor,
                colorScheme: ColorScheme.fromSeed(
                  seedColor: value.themedColors.accentColor,
                  brightness:
                      value.isDarkTheme ? Brightness.dark : Brightness.light,
                ),
              ),
              showPerformanceOverlay: showPerformanceOverlay,
              themeMode: value.isDarkTheme ? ThemeMode.dark : ThemeMode.light,
              title: title,
              debugShowCheckedModeBanner: debugShowCheckedModeBanner,
            );
          },
        );
      }),
    );
  }
}
