import 'package:flutter/foundation.dart';
import 'package:flutter/gestures.dart';
import 'package:flutter/material.dart';
import 'package:flutter/rendering.dart';
import 'package:flutter/services.dart';
import 'package:flutter_localizations/flutter_localizations.dart';

import 'themed_hover_button.dart';

const double _kMinTileWidth = 80.0;
const double _kMaxTileWidth = 240.0;
const double _kTileHeight = 34.0;
const double _kButtonWidth = 32.0;
typedef ThemedTab = Tab;

enum CloseButtonVisibilityMode {
  /// The close button will never be visible
  never,

  /// The close button will always be visible
  always,

  /// The close button will only be shown on hover
  onHover,
}

/// Determines how the tab sizes itself
enum TabWidthBehavior {
  /// The tab will fit its content
  sizeToContent,

  /// If not scrollable, the tabs will have the same size
  equal,

  /// If not selected, the [Tab]'s text is hidden. The tab will fit its content
  compact,
}

/// The TabView control is a way to display a set of tabs and their respective
/// content. TabViews are useful for displaying several pages (or documents) of
/// content while giving a user the capability to rearrange, open, or close new
/// tabs.
///
/// ![TabView Preview](https://docs.microsoft.com/en-us/windows/apps/design/controls/images/tabview/tab-introduction.png)
///
/// There must be enough space to render the tabview.
///
/// See also:
///
///   * [NavigationView], control provides top-level navigation for your app.
///   * <https://docs.microsoft.com/en-us/windows/apps/design/controls/tab-view>
class ThemedTabView extends StatefulWidget {
  /// Creates a tab view.
  ///
  /// [tabs] must have the same length as [bodies]
  ///
  /// [maxTabWidth] must be non-negative
  const ThemedTabView({
    Key? key,
    required this.currentIndex,
    this.onChanged,
    required this.tabs,
    this.onNewPressed,
    this.addIconData = Icons.add,
    this.shortcutsEnabled = true,
    this.onReorder,
    this.showScrollButtons = true,
    this.wheelScroll = false,
    this.scrollController,
    this.minTabWidth = _kMinTileWidth,
    this.maxTabWidth = _kMaxTileWidth,
    this.closeButtonVisibility = CloseButtonVisibilityMode.always,
    this.tabWidthBehavior = TabWidthBehavior.equal,
    this.header,
    this.footer,
  }) : super(key: key);

  /// The index of the tab to be displayed
  final int currentIndex;

  /// Whether another tab was requested to be displayed
  final ValueChanged<int>? onChanged;

  /// The tabs to be displayed. This must have the same
  /// length of [bodies]
  final List<Tab> tabs;

  /// Called when the new button is pressed or when the
  /// shortcut `Ctrl + T` is executed.
  ///
  /// If null, the new button won't be displayed
  final VoidCallback? onNewPressed;

  /// The icon of the new button
  final IconData addIconData;

  /// Whether the following shortcuts are enabled:
  ///
  /// - Ctrl + T to create a new tab
  /// - Ctrl + F4 or Ctrl + W to close the current tab
  /// - `Ctrl+1` to `Ctrl+8` to navigate through tabs
  /// - `Ctrl+9` to navigate to the last tab
  final bool shortcutsEnabled;

  /// Called when the tabs are reordered. If null,
  /// reordering is disabled. It's disabled by default.
  final ReorderCallback? onReorder;

  /// The min width a tab can have. Must not be negative.
  ///
  /// Default to 80 logical pixels
  final double minTabWidth;

  /// The max width a tab can have. Must not be negative.
  ///
  /// Defaults to 240 logical pixels
  final double maxTabWidth;

  /// Whether the buttons that scroll forward or backward
  /// should be displayed, if necessary. Defaults to true
  final bool showScrollButtons;

  /// The [ScrollPosController] used to move tabview to right and left when the
  /// tabs don't fit the available horizontal space.
  ///
  /// If null, a [ScrollPosController] is created internally.
  final ScrollPosController? scrollController;

  // TODO: remove this property when https://github.com/flutter/flutter/issues/75180
  // is fixed
  /// Indicate if the mouse wheel should scroll the TabView
  ///
  /// Defaults to `false`.
  ///
  final bool wheelScroll;

  /// Indicates the close button visibility mode
  final CloseButtonVisibilityMode closeButtonVisibility;

  /// Indicates how a tab will size itself
  final TabWidthBehavior tabWidthBehavior;

  /// Displayed before all the tabs and buttons.
  ///
  /// Usually a [Text]
  final Widget? header;

  /// Displayed after all the tabs and buttons.
  ///
  /// Usually a [Text]
  final Widget? footer;

  /// Whenever the new button should be displayed.
  bool get showNewButton => onNewPressed != null;

  /// Whether reordering is enabled or not. To enable it,
  /// make sure [widget.onReorder] is not null.
  bool get isReorderEnabled => onReorder != null;

  @override
  State<StatefulWidget> createState() => _TabViewState();

  @override
  void debugFillProperties(DiagnosticPropertiesBuilder properties) {
    super.debugFillProperties(properties);
    properties.add(IntProperty('currentIndex', currentIndex));
    properties.add(FlagProperty(
      'showNewButton',
      value: showNewButton,
      ifFalse: 'no new button',
    ));
    properties.add(IconDataProperty('addIconData', addIconData));
    properties.add(ObjectFlagProperty(
      'onChanged',
      onChanged,
      ifNull: 'disabled',
    ));
    properties.add(ObjectFlagProperty(
      'onNewPressed',
      onNewPressed,
      ifNull: 'no new button',
    ));
    properties.add(IntProperty('tabs', tabs.length));
    properties.add(FlagProperty(
      'reorderEnabled',
      value: isReorderEnabled,
      ifFalse: 'reorder disabled',
    ));
    properties.add(FlagProperty(
      'showScrollButtons',
      value: showScrollButtons,
      ifFalse: 'hide scroll buttons',
    ));
    properties.add(EnumProperty('closeButtonVisibility', closeButtonVisibility,
        defaultValue: CloseButtonVisibilityMode.always));
  }
}

class _TabViewState extends State<ThemedTabView> {
  late ScrollPosController scrollController;

  @override
  void initState() {
    super.initState();
    scrollController = widget.scrollController ??
        ScrollPosController(
          itemCount: widget.tabs.length,
          animationDuration: const Duration(milliseconds: 100),
        );
    scrollController.itemCount = widget.tabs.length;
    scrollController.addListener(_handleScrollUpdate);
  }

  void _handleScrollUpdate() {
    if (mounted) setState(() {});
  }

  @override
  void didUpdateWidget(ThemedTabView oldWidget) {
    super.didUpdateWidget(oldWidget);
    if (widget.tabs.length != oldWidget.tabs.length) {
      scrollController.itemCount = widget.tabs.length;
    }
    if (widget.currentIndex != oldWidget.currentIndex) {
      scrollController.scrollToItem(widget.currentIndex, center: false);
    }
  }

  @override
  void dispose() {
    if (widget.scrollController == null) {
      // only dispose the local controller
      scrollController.dispose();
    }
    super.dispose();
  }

  Widget _tabBuilder(
    BuildContext context,
    int index,
    double preferredTabWidth,
  ) {
    final Tab tab = widget.tabs[index];
    final tabWidget = _Tab(
      tab,
      key: ValueKey<int>(index),
      reorderIndex: widget.isReorderEnabled ? index : null,
      selected: index == widget.currentIndex,
      onPressed:
          widget.onChanged == null ? null : () => widget.onChanged!(index),
      visibilityMode: widget.closeButtonVisibility,
      tabWidthBehavior: widget.tabWidthBehavior,
    );
    final Widget child = GestureDetector(
      onTertiaryTapUp: (_) => tab.onClosed?.call(),
      child: Row(mainAxisSize: MainAxisSize.min, children: [
        if (widget.tabWidthBehavior == TabWidthBehavior.equal)
          Expanded(child: tabWidget)
        else
          Flexible(child: tabWidget),
      ]),
    );
    final minWidth = () {
      switch (widget.tabWidthBehavior) {
        case TabWidthBehavior.sizeToContent:
        case TabWidthBehavior.compact:
          return null;
        default:
          return preferredTabWidth;
      }
    }();
    if (minWidth == null) {
      return KeyedSubtree(
        key: ValueKey<Tab>(tab),
        child: child,
      );
    }
    return AnimatedContainer(
      key: ValueKey<Tab>(tab),
      constraints: BoxConstraints(maxWidth: minWidth, minWidth: minWidth),
      duration: Duration(milliseconds: 1000),
      child: child,
    );
  }

  Widget _buttonTabBuilder(
    BuildContext context,
    Widget icon,
    VoidCallback? onPressed,
    String tooltip,
  ) {
    final item = SizedBox(
      width: _kButtonWidth,
      height: 24.0,
      child: IconButton(
        icon: Center(child: icon),
        onPressed: onPressed,
        style: ButtonStyle(),
      ),
    );
    if (onPressed == null) return item;
    return Tooltip(message: tooltip, child: item);
  }

  @override
  Widget build(BuildContext context) {
    final direction = Directionality.of(context);

    final headerFooterTextStyle = const TextStyle();

    Widget tabBar = Column(
      children: [
        ScrollConfiguration(
          behavior: const _TabViewScrollBehavior(),
          child: Container(
            margin: const EdgeInsets.only(top: 4.5),
            padding: const EdgeInsets.only(left: 8),
            height: _kTileHeight,
            width: double.infinity,
            child: Row(children: [
              if (widget.header != null)
                Padding(
                  padding: const EdgeInsets.only(right: 12.0),
                  child: DefaultTextStyle(
                    style: headerFooterTextStyle,
                    child: widget.header!,
                  ),
                ),
              Expanded(
                child: LayoutBuilder(builder: (context, consts) {
                  final width = consts.biggest.width;
                  assert(
                    width.isFinite,
                    'You can only create a TabView in a box with defined width',
                  );

                  final double preferredTabWidth =
                      ((width - (widget.showNewButton ? _kButtonWidth : 0)) /
                              widget.tabs.length)
                          .clamp(widget.minTabWidth, widget.maxTabWidth);

                  final Widget listView = Listener(
                    onPointerSignal: widget.wheelScroll
                        ? (PointerSignalEvent e) {
                            if (e is PointerScrollEvent) {
                              if (e.scrollDelta.dy > 0) {
                                scrollController.forward(
                                  align: false,
                                  animate: false,
                                );
                              } else {
                                scrollController.backward(
                                  align: false,
                                  animate: false,
                                );
                              }
                            }
                          }
                        : null,
                    child: Localizations.override(
                      context: context,
                      delegates: const [
                        GlobalMaterialLocalizations.delegate,
                      ],
                      child: ReorderableListView.builder(
                        buildDefaultDragHandles: false,
                        shrinkWrap: true,
                        scrollDirection: Axis.horizontal,
                        scrollController: scrollController,
                        onReorder: (i, ii) {
                          widget.onReorder?.call(i, ii);
                        },
                        itemCount: widget.tabs.length,
                        proxyDecorator: (child, index, animation) {
                          return child;
                        },
                        itemBuilder: (context, index) {
                          return _tabBuilder(context, index, preferredTabWidth);
                        },
                      ),
                    ),
                  );

                  bool scrollable = preferredTabWidth * widget.tabs.length >
                      width - (widget.showNewButton ? _kButtonWidth : 0);

                  final bool showScrollButtons =
                      widget.showScrollButtons && scrollable;
                  final backwardButton = Padding(
                    padding: const EdgeInsetsDirectional.only(
                      start: 8.0,
                      end: 3.0,
                      bottom: 3.0,
                    ),
                    child: _buttonTabBuilder(
                      context,
                      const Icon(
                          IconData(
                            0xedd9,
                          ),
                          size: 8),
                      !scrollController.canBackward
                          ? () {
                              if (direction == TextDirection.ltr) {
                                scrollController.backward();
                              } else {
                                scrollController.forward();
                              }
                            }
                          : null,
                      'Scroll Tab Backward',
                    ),
                  );

                  final forwardButton = Padding(
                    padding: const EdgeInsetsDirectional.only(
                      start: 3.0,
                      end: 8.0,
                      bottom: 3.0,
                    ),
                    child: _buttonTabBuilder(
                      context,
                      const Icon(
                          IconData(
                            0xedda,
                          ),
                          size: 8),
                      !scrollController.canForward
                          ? () {
                              if (direction == TextDirection.ltr) {
                                scrollController.forward();
                              } else {
                                scrollController.backward();
                              }
                            }
                          : null,
                      'Scroll Tab Forward',
                    ),
                  );

                  return Row(children: [
                    if (showScrollButtons)
                      direction == TextDirection.ltr
                          ? backwardButton
                          : forwardButton,
                    if (scrollable)
                      Expanded(child: listView)
                    else
                      Flexible(child: listView),
                    if (showScrollButtons)
                      direction == TextDirection.ltr
                          ? forwardButton
                          : backwardButton,
                    if (widget.showNewButton)
                      Padding(
                        padding: const EdgeInsetsDirectional.only(
                          start: 3.0,
                          bottom: 3.0,
                        ),
                        child: _buttonTabBuilder(
                          context,
                          Icon(widget.addIconData, size: 12.0),
                          widget.onNewPressed!,
                          'New Tab',
                        ),
                      ),
                  ]);
                }),
              ),
              if (widget.footer != null)
                Padding(
                  padding: const EdgeInsets.only(left: 12.0),
                  child: DefaultTextStyle(
                    style: headerFooterTextStyle,
                    child: widget.footer!,
                  ),
                ),
            ]),
          ),
        ),
        if (widget.tabs.isNotEmpty)
          Expanded(
            child: IndexedStack(
              index: widget.currentIndex,
              children: widget.tabs.map((e) => e.body).toList(),
            ),
          ),
      ],
    );
    if (widget.shortcutsEnabled) {
      void onClosePressed() {
        widget.tabs[widget.currentIndex].onClosed?.call();
      }

      return FocusScope(
        autofocus: true,
        child: CallbackShortcuts(
          bindings: {
            const SingleActivator(LogicalKeyboardKey.f4, control: true):
                onClosePressed,
            const SingleActivator(LogicalKeyboardKey.keyW, control: true):
                onClosePressed,
            const SingleActivator(LogicalKeyboardKey.keyT, control: true): () {
              widget.onNewPressed?.call();
            },
            ...Map.fromIterable(
              List<int>.generate(9, (index) => index),
              key: (i) {
                final digits = [
                  LogicalKeyboardKey.digit1,
                  LogicalKeyboardKey.digit2,
                  LogicalKeyboardKey.digit3,
                  LogicalKeyboardKey.digit4,
                  LogicalKeyboardKey.digit5,
                  LogicalKeyboardKey.digit6,
                  LogicalKeyboardKey.digit7,
                  LogicalKeyboardKey.digit8,
                  LogicalKeyboardKey.digit9,
                ];
                return SingleActivator(digits[i], control: true);
              },
              value: (index) {
                return () {
                  // If it's the last, move to the last tab
                  if (index == 8) {
                    widget.onChanged?.call(widget.tabs.length - 1);
                  } else {
                    if (widget.tabs.length - 1 >= index) {
                      widget.onChanged?.call(index);
                    }
                  }
                };
              },
            ),
          },
          child: tabBar,
        ),
      );
    }
    return tabBar;
  }
}

/// Represents a single tab within a [TabView].
class Tab {
  /// Creates a tab.
  const Tab({
    this.key,
    this.icon = const SizedBox.shrink(),
    required this.text,
    required this.body,
    this.closeIcon = const IconData(
      0xe8bb,
    ),
    this.onClosed,
    this.semanticLabel,
  });

  final Key? key;

  /// the IconSource to be displayed within the tab.
  ///
  /// Usually an [Icon] widget
  final Widget? icon;

  /// The content that appears inside the tab strip to represent the tab.
  ///
  /// Usually a [Text] widget
  final Widget text;

  /// The close icon of the tab. Usually an [IconButton] widget
  final IconData? closeIcon;

  /// Called when clicking x-to-close button or when thec`Ctrl + T` or
  /// `Ctrl + F4` is executed
  ///
  /// If null, the tab is not closeable
  final VoidCallback? onClosed;

  final String? semanticLabel;

  /// The body of the view attached to this tab
  final Widget body;
}

class _Tab extends StatefulWidget {
  const _Tab(
    this.tab, {
    Key? key,
    this.onPressed,
    required this.selected,
    this.reorderIndex,
    this.animationDuration = Duration.zero,
    this.animationCurve = Curves.linear,
    required this.visibilityMode,
    required this.tabWidthBehavior,
  }) : super(key: key);

  final Tab tab;
  final bool selected;
  final VoidCallback? onPressed;
  final int? reorderIndex;
  final Duration animationDuration;
  final Curve animationCurve;
  final CloseButtonVisibilityMode visibilityMode;
  final TabWidthBehavior tabWidthBehavior;

  @override
  __TabState createState() => __TabState();
}

class __TabState extends State<_Tab>
    with SingleTickerProviderStateMixin, AutomaticKeepAliveClientMixin {
  late AnimationController _controller;

  @override
  void initState() {
    super.initState();
    _controller = AnimationController(
      vsync: this,
      duration: widget.animationDuration,
    );
    _controller.forward();
  }

  @override
  void dispose() {
    _controller.dispose();
    super.dispose();
  }

  @override
  void didUpdateWidget(_Tab oldWidget) {
    super.didUpdateWidget(oldWidget);
    _controller.duration = oldWidget.animationDuration;
  }

  @override
  Widget build(BuildContext context) {
    super.build(context);

    // The text of the tab, if a [Text] widget is used
    final String? text = () {
      if (widget.tab.text is Text) {
        return (widget.tab.text as Text).data ??
            (widget.tab.text as Text).textSpan?.toPlainText();
      } else if (widget.tab.text is RichText) {
        return (widget.tab.text as RichText).text.toPlainText();
      }
    }();
    return ThemedHoverButton(
      key: widget.tab.key,
      semanticLabel: widget.tab.semanticLabel ?? text,
      onPressed: widget.onPressed,
      builder: (context, states) {
        final foregroundColor =
            ButtonState.resolveWith((states) {}).resolve(states);
        const borderRadius = BorderRadius.vertical(top: Radius.circular(6));
        Widget child = Container(
          height: _kTileHeight,
          constraints: widget.tabWidthBehavior == TabWidthBehavior.sizeToContent
              ? null
              : const BoxConstraints(maxWidth: _kMaxTileWidth),
          padding: const EdgeInsetsDirectional.only(
            start: 8,
            top: 3,
            end: 4,
            bottom: 3,
          ),
          decoration: BoxDecoration(
            borderRadius: borderRadius,
            color: widget.selected
                ? null
                : Theme.of(context).brightness == Brightness.dark
                    ? Colors.grey[600]
                    : Colors.grey[200],
          ),
          child: () {
            final result = ClipRect(
              child: IconTheme.merge(
                data: IconThemeData(
                  color: foregroundColor,
                  size: 16.0,
                ),
                child: Row(mainAxisSize: MainAxisSize.min, children: [
                  if (widget.tab.icon != null)
                    Padding(
                      padding: const EdgeInsetsDirectional.only(end: 10.0),
                      child: widget.tab.icon!,
                    ),
                  if (widget.tabWidthBehavior != TabWidthBehavior.compact ||
                      (widget.tabWidthBehavior == TabWidthBehavior.compact &&
                          widget.selected))
                    Flexible(
                      fit: widget.tabWidthBehavior == TabWidthBehavior.equal
                          ? FlexFit.tight
                          : FlexFit.loose,
                      child: Padding(
                        padding: const EdgeInsetsDirectional.only(end: 4.0),
                        child: DefaultTextStyle.merge(
                          softWrap: false,
                          maxLines: 1,
                          overflow: TextOverflow.clip,
                          style: const TextStyle(fontSize: 12.0),
                          child: widget.tab.text,
                        ),
                      ),
                    ),
                  if (widget.tab.closeIcon != null &&
                      (widget.visibilityMode ==
                              CloseButtonVisibilityMode.always ||
                          (widget.visibilityMode ==
                                  CloseButtonVisibilityMode.onHover &&
                              states.isHovering)))
                    Padding(
                      padding: const EdgeInsetsDirectional.only(start: 4.0),
                      child: Tooltip(
                        message: 'Close Tab',
                        child: SizedBox(
                          height: 24.0,
                          width: 32.0,
                          child: IconButton(
                            icon: Icon(widget.tab.closeIcon),
                            onPressed: widget.tab.onClosed,
                          ),
                        ),
                      ),
                    ),
                ]),
              ),
            );
            if (widget.reorderIndex != null) {
              return ReorderableDragStartListener(
                index: widget.reorderIndex!,
                child: result,
              );
            }
            return result;
          }(),
        );
        if (text != null) {
          child = Tooltip(
            message: text,
            child: child,
          );
        }
        if (widget.selected) {
          child = CustomPaint(
            willChange: false,
            painter: _TabPainter(
              Theme.of(context).brightness == Brightness.light
                  ? Colors.white
                  : Colors.black,
            ),
            child: child,
          );
        }
        return Semantics(
          selected: widget.selected,
          focusable: true,
          focused: states.isFocused,
          child: child,
        );
      },
    );
  }

  @override
  bool get wantKeepAlive => true;
}

class _TabPainter extends CustomPainter {
  final Color color;

  const _TabPainter(this.color);

  @override
  void paint(Canvas canvas, Size size) {
    final path = Path();
    const radius = 6.0;
    path
      ..moveTo(-radius, size.height)
      ..quadraticBezierTo(0, size.height, 0, size.height - radius)
      ..lineTo(0, radius)
      ..quadraticBezierTo(0, 0, radius, 0)
      ..lineTo(size.width - radius, 0)
      ..quadraticBezierTo(size.width, 0, size.width, radius)
      ..lineTo(size.width, size.height - radius)
      ..quadraticBezierTo(
        size.width,
        size.height,
        size.width + radius,
        size.height,
      );
    canvas.drawPath(path, Paint()..color = color);
  }

  @override
  bool shouldRepaint(_TabPainter oldDelegate) => color != oldDelegate.color;

  @override
  bool shouldRebuildSemantics(_TabPainter oldDelegate) => false;
}

class _TabViewScrollBehavior extends ScrollBehavior {
  const _TabViewScrollBehavior();

  @override
  Widget buildScrollbar(context, child, details) {
    return child;
  }
}

const _kDefaultAnimationDuration = Duration(milliseconds: 300);
const _kDefaultAnimationCurve = Curves.ease;

/// Controls a scrollable widget.
///
/// This class extends [ScrollController] and adds functionality to it.
/// A ScrollPosController allow you to place an item on a specific area
/// of the screen.
class ScrollPosController extends ScrollController {
  int itemCount;
  bool animate;
  Duration animationDuration;
  Curve animationCurve;

  ScrollPosController({
    this.itemCount = 0,
    this.animate = true,
    this.animationDuration = _kDefaultAnimationDuration,
    this.animationCurve = _kDefaultAnimationCurve,
    double initialScrollOffset = 0.0,
    bool keepScrollOffset = true,
    String? debugLabel,
  }) : super(
          initialScrollOffset: initialScrollOffset,
          keepScrollOffset: keepScrollOffset,
          debugLabel: debugLabel,
        );

  @visibleForTesting
  double get inside => positions.isNotEmpty ? position.extentInside : 0;

  @visibleForTesting
  double get max => positions.isNotEmpty ? position.maxScrollExtent : 0;

  /// total returns the total of the inside part and the scrollable part.
  double get total => inside + max;

  /// scrollPerItem returns the quantity of scroll needed for travel
  /// the height of one item.
  double get scrollPerItem => total / itemCount;

  /// visibleItems returns the quantity of items visible in the screen.
  double get visibleItems => inside / scrollPerItem;

  /// canScroll indicate if this widget has scrollable content
  bool get canScroll => max > 0;

  /// canForward indicate if the scrollbar is at the end position.
  bool get canForward => positions.isNotEmpty ? offset == max : false;

  /// canBackward indicate if the scrollbar is at the start position.
  bool get canBackward => positions.isNotEmpty ? offset == 0 : false;

  /// scrollTop will move the scrollbar to the top (first item).
  void scrollToStart({bool? animate}) {
    _scrollToPos(0, animate: animate);
  }

  /// scrollBottom will move the scrollbar to the bottom (last item).
  void scrollToEnd({bool? animate}) {
    _scrollToPos(max, animate: animate);
  }

  /// scrollToItem will move the scrollbar to make the item
  /// (defined by its index) visible on the screen.
  /// If the item is already fully visible in the screen, the scrollbar
  /// is not moved.
  void scrollToItem(int index, {bool? animate, bool center = false}) {
    _scrollOffItem(index, animate: animate, center: center);
  }

  /// backward will move the scrollbar with a distance of one item in
  /// the bottom direction.
  /// If value of align is `true`, the last visible item will be aligned.
  void forward({bool align = true, bool? animate}) {
    if (align) {
      final lastIdx = ((offset + inside) / scrollPerItem - 1).floor() + 1;
      if (lastIdx >= itemCount) {
        return;
      }
      scrollToItem(lastIdx, animate: animate);
    } else {
      _scrollToPos(offset + scrollPerItem, animate: animate);
    }
  }

  /// backward will move the scrollbar with a distance of one item in
  /// the top direction.
  /// If value of align is `true`, the first visible item will be aligned.
  void backward({bool align = true, bool? animate}) {
    if (align) {
      final firstIdx = (offset / scrollPerItem - 1).ceil();
      if (firstIdx < 0) {
        return;
      }
      scrollToItem(firstIdx, animate: animate);
    } else {
      _scrollToPos(offset - scrollPerItem, animate: animate);
    }
  }

  /// scrollOffItemCenter return the offset when the item is placed at
  /// the center of the screen.
  double scrollOffItemCenter(int index) {
    return (index + 1) * scrollPerItem - inside / 2;
  }

  /// scrollOfItemBottom return the offset when the item is placed at
  /// the bottom of the screen.
  double scrollOffItemEnd(int index) {
    return (index + 1) * scrollPerItem - inside;
  }

  /// scrollOffItemTop return the offset when the item is placed at
  /// the top of the screen.
  double scrollOffItemStart(int index) {
    return index * scrollPerItem;
  }

  void _scrollToPos(double offset, {bool? animate}) {
    if (((animate != null && animate) || (animate == null && this.animate)) &&
        animationDuration != Duration.zero) {
      animateTo(
        offset.clamp(0, max),
        duration: animationDuration,
        curve: animationCurve,
      );
    } else {
      jumpTo(offset.clamp(0, max));
    }
  }

  void _scrollOffItem(int index, {bool? animate, bool center = false}) {
    final centerVal = scrollOffItemCenter(index);
    final endVal = scrollOffItemEnd(index);
    final startVal = scrollOffItemStart(index);

    if (center) {
      _scrollToPos(centerVal, animate: animate);
    } else if (offset < endVal) {
      _scrollToPos(endVal, animate: animate);
    } else if (offset > startVal) {
      _scrollToPos(startVal, animate: animate);
    }
  }
}
